########################################################################
#
# Copyright (C) 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''tiling utilities'''

from math import ceil, log

class TMSTiles(object):
    ###
    ### This class is based upon https://github.com/OSGeo/gdal/blob/master/swig/python/gdal-utils/osgeo_utils/gdal2tiles.py
    ### written by:       Klokan Petr Pridal (klokan at klokan dot cz)
    ###                   Even Rouault (even dot rouault at spatialys.com)
    ###                   Idan Miara (idan@miara.com)
    ### It is a small subset of the original source code, and it is highly modified to only
    ### handle raster profile tiling. The original source code has many useful details and
    ### supports mercator and geodetic data. The subset of gdal2tiles.py used here is less
    ### than 3% of the original file (42 lines out of 1697). While this is not considered
    ### a substantial portion, it was invaluable for understanding the TMS tile layout.
    ###
    def __init__( self, dimension, title='' ):
        '''
        Construct a tiling object which will provide the coordinates of tiles, the size to
        be read, the size of the scaled tile and the write offset within the :code:`tile_size`
        output tile.

        Parameters
        ----------
        dimension: list or tuple
            should contain at least two elements (X,Y)
        '''

        def calculate_indexes( tz, result={ } ):

            ti = 0
            tile_details = []
            tminx, tminy, tmaxx, tmaxy = self.tminmax[tz]

            for ty in range(tmaxy, tminy - 1, -1):
                for tx in range(tminx, tmaxx + 1):
                    ti += 1
                    ytile = ty
                    index = ( tz, tx, ty )

                    tsize = int(self.tsize[tz])   # tile_size in raster coordinates for actual zoom
                    xsize = self.__dimension[0]
                    ysize = self.__dimension[1]
                    querysize = self.__tile_size

                    rx = tx * tsize
                    rxsize = 0
                    if tx == tmaxx:
                        rxsize = xsize % tsize
                    if rxsize == 0:
                        rxsize = tsize

                    ry = ty * tsize
                    rysize = 0
                    if ty == tmaxy:
                        rysize = ysize % tsize
                    if rysize == 0:
                        rysize = tsize

                    wx, wy = 0, 0
                    wxsize = int(rxsize / float(tsize) * self.__tile_size)
                    wysize = int(rysize / float(tsize) * self.__tile_size)

                    ry = ysize - (ty * tsize) - rysize
                    if wysize != self.__tile_size:
                        wy = self.__tile_size - wysize

                    result[index] = dict( src=dict( idx=(rx, ry), dim=(rxsize, rysize),
                                                    t=(( 0, 0 if rysize == tsize else tsize - rysize ), tsize ) ),
                                          dst=dict( idx=(wx, wy), dim=(wxsize, wysize) ) )
            return result

        self.__title = title
        self.__tile_size = 256
        self.__dimension = dimension
        self.__zoom = (0, max( 0, int( max( ceil(log(dimension[0] / float(self.__tile_size), 2 )),
                                            ceil( log(dimension[1] / float(self.__tile_size), 2 )) ) ) ) )
        native_zoom = self.__zoom[1]

        # Generate table with min max tile coordinates for all zoomlevels
        self.tminmax = list(range(self.__zoom[0], self.__zoom[1] + 1))
        self.tsize = list(range(self.__zoom[0], self.__zoom[1] + 1))
        for tz in range(self.__zoom[0], self.__zoom[1] + 1):
            tsize = 2.0**(native_zoom - tz) * self.__tile_size
            tminx, tminy = 0, 0
            tmaxx = int(ceil(dimension[0] / tsize)) - 1
            tmaxy = int(ceil(dimension[1] / tsize)) - 1
            self.tsize[tz] = ceil(tsize)
            self.tminmax[tz] = (tminx, tminy, tmaxx, tmaxy)

        self.__units_per_pixel = { }
        for z in range(self.__zoom[0], self.__zoom[1] + 1):
            self.__units_per_pixel[z] = 2**(self.__zoom[1] - z)

        self.__tile_details = { }
        for z in range(self.__zoom[0], self.__zoom[1] + 1):
            calculate_indexes( z, self.__tile_details )

    def profile( self ):
        return "raster"

    def dim( self ):
        return self.__dimension

    def tile_size( self ):
        return self.__tile_size

    def tile( self, z, x=None, y=None ):
        try:
            if (type(z) == tuple or type(z) == list) and len(z) == 3 and x is None and y is None:
                return self.__tile_details[ (int(z[0]),int(z[1]),int(z[2])) ]
            elif type(x) is not None and type(y) is not None:
                return self.__tile_details[ (int(z),int(x),int(y)) ]
            else:
                raise RuntimeError( 'tile requires 3 integer parameters or 1 tuple[3] parameter' )
        except KeyError:
            return None

    def zoom_levels( self, reverse=False ):
        return sorted( list(self.__units_per_pixel.keys( )), reverse=reverse )

    def units_per_pixel( self, zoom_level ):
        if zoom_level not in self.__units_per_pixel:
            raise RuntimeError(f'''{zoom_level} is not an existing zoom level''')
        return self.__units_per_pixel[zoom_level]

    def __str__( self ):
        return f'''<?xml version="1.0" encoding="utf-8"?>
    <TileMap version="1.0.0" tilemapservice="http://tms.osgeo.org/1.0.0">
      <Title>{self.__title}</Title>
      <Abstract></Abstract>
      <SRS></SRS>
      <BoundingBox minx="0.0000" miny="0.0000" maxx="{self.__dimension[0]}.0000" maxy="{self.__dimension[1]}.0000"/>
      <Origin x="0.0000" y="0.0000"/>
      <TileFormat width="256" height="256" mime-type="image/png" extension="png"/>
      <TileSets profile="raster">''' + "\n" + '\n'.join( map( lambda z: f'''        <TileSet href="{z}" units-per-pixel="{self.units_per_pixel(z)}.0000" order="{z}"/>''',
                                                              self.zoom_levels( ) ) ) + f'''
      </TileSets>
    </TileMap>
'''



########################################################################
#
# Copyright (C) 2021,2022,2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''General utility functions used by the ``casagui`` tools and applications.'''

import inspect
import itertools
import urllib.request
import urllib.error
import sys

from itertools import groupby, chain
from socket import socket
from os import path as __path
from ._ResourceManager import _ResourceManager
from ._logging import get_logger
from ._regions import polygon_indexes
from ._docenum import DocEnum

from astropy import units as u
from regions import PixCoord
from regions import RectanglePixelRegion, PolygonPixelRegion

try:
    from casatools import regionmanager

    __have_casatools = True
except ImportError:
    __have_casatools = False

logger = get_logger()

from ._conversion import pack_arrays
from ._conversion import strip_arrays
from ._conversion import serialize, deserialize
from ._static import static_vars, static_dir
from ._image import image_as_mime
from ._tiles import TMSTiles

@static_vars(mgr=None)
def resource_manager( ):
    if resource_manager.mgr is None:
        resource_manager.mgr = _ResourceManager( )
    return resource_manager.mgr

def reset_resource_manager( ):
    if resource_manager.mgr is not None:
        resource_manager.mgr = _ResourceManager( )

def path_to_url(path):
    '''Convert a single filesystem path to a URL.

    If the string specified in the ``path`` parameter exists. It is turned into a
    fully qualified path and converted to a URL and returned. If ``path`` does not
    exist, ``path`` is returned unchanged.

    Parameters
    ----------
    path: str
        path to be checked and expanded

    Returns
    -------
    str
        ``path`` converted to a URL if ``path`` exists, otherwise ``path`` unchanged
    '''
    return "file://" + __path.abspath(path) if __path.exists(path) else path


def find_ws_address(address='127.0.0.1'):
    '''Find free port on ``address`` network and return a tuple with ``address`` and port number

    This function uses the low level socket function to find a free port and return
    a tuple representing the address plus port number.

    Parameters
    ----------
    address: str
        network to be probed for an available port

    Returns
    -------
    tuple of str and int
        network address (`str`) and port number (`int`)
    '''
    sock = socket()
    sock.bind((address, 0))
    result = sock.getsockname()
    sock.close()
    return result


def partition(pred, iterable):
    '''Split ``iterable`` into two lists based on ``pred`` predicate.
    '''
    trues = []
    falses = []
    for item in iterable:
        if pred(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses


def error_msg(*args, **kwargs):
    '''standard method for reporting errors which do not result in aborting out of python

    This function takes the standard set of arguments that the python ``print`` function takes.
    The primary difference is that the output will go to ``stderr`` and perhaps other error
    logs.
    '''
    print(*args, file=sys.stderr, **kwargs)


@static_vars(msgs=dict(
    casatools='{package} is not available so interactive clean and plotants will not work',
    casatasks='{package} is not available so interactive clean will not work'
), reported={})
def warn_import(package):
    '''standard method for reporting (optional) package import failure

    Parameters
    ----------
    package: str
        name of a package whose attempted import failed
    '''
    if package not in warn_import.reported:
        warn_import.reported[package] = True
        error_msg("warning, %s" % warn_import.msgs[package].format(package=package))


@static_vars(url='http://clients3.google.com/generate_204')
def have_network():
    '''check to see if an active network with general internet connectivity
    is available. returns ``True`` if we have internet connectivity and
    ``False`` if we do not.
    '''
    ###
    ### see: https://stackoverflow.com/questions/50558000/test-internet-connection-for-python3
    ###
    try:
        with urllib.request.urlopen(have_network.url) as response:
            return response.status == 204
    except urllib.error.HTTPError:
        ### http error
        return False
    except urllib.error.ContentTooShortError:
        return False
    except urllib.error.URLError:
        return False
    except Exception:
        return False


def ranges(iterable, order=sorted, key=lambda x: x):
    '''collect elements of ``iterable`` into tuple ranges where each tuple represents
    a concesecutive range within the iterable. ``key`` can be used to provide ranges
    for other objects where ``key(element)`` returns the key to be used for sorting
    into ranges'''
    for a, b in groupby(enumerate(
            order(iterable, key=key) if 'key' in inspect.getfullargspec(order).kwonlyargs else order(iterable)),
                        lambda pair: key(pair[1]) - pair[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def contiguous_ranges(iterable, order=sorted, key=lambda x: x):
    '''split iterable into contiguous index sequence, i.e. reduce consecutive index runs
    to a tuple containing the first and last. ``iterable`` can be a sequence of values that
    can be subtracted or a list of more complex values where the function ``key`` returns
    the key to be used for ordering. ``iterable`` is ordered by ``order`` which by default
    sorts ``iterable``.'''
    unique = list(set(iterable))
    for a, b in groupby(
            enumerate(order(unique, key=key) if 'key' in inspect.getfullargspec(order).kwonlyargs else order(unique)),
            lambda pair: key(pair[1]) - pair[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def expand_range_incl(r):
    '''expand the tuple supplied as ``r`` into a range which *includes* the first
    and last element ``r``'''
    if r[0] == r[1]:
        return [0]
    else:
        return range(r[0], r[1] + 1)


def __get_wcs_object(csys: 'image.coordsys', naxis=2) -> 'image.coordsys()':
    """Generate world coordinate object from coordinate system.

    Returns:
        w: world coordinate object
    """
    try:
        import numpy as np
        from astropy.wcs import WCS

        w = WCS(naxis=naxis)

        rad_to_deg = 180/np.pi
        w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
        w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
        w.wcs.crval = csys.referencevalue()['numeric'][0:2] * rad_to_deg
        w.wcs.ctype = ['RA---SIN', 'DEC--SIN']

        return w

    except ImportError as error:
        print('Error importing module:: ' + str(error))

    except Exception as error:
        print(error)

def __index_to_stokes(index: int):
    """Convert stokes axis index to alphabetic value.

    Args:
        index (int): enumerated index defining stokes value.

    Returns:
        str: String indicating stokes value. 
    """
    STOKES_MAP = ['I', 'Q', 'U', 'V']
  
    try:
        return [STOKES_MAP[i] for i in index]
    except TypeError:
        return STOKES_MAP[index]

def index_to_stokes(index: int):
    """Convert stokes axis index to alphabetic value.

    Args:
        index (int): enumerated index defining stokes value.

    Returns:
        str: String indicating stokes value. 
    """
    STOKES_MAP = ['I', 'Q', 'U', 'V']
  
    try:
        return [STOKES_MAP[i] for i in index]
    except TypeError:
        return STOKES_MAP[index]

def __get_center_pixels(params: dict):
    """Get center pixel offset value from dictionary.

    Args:
        params (dict): Submask from mask grouping used by itertools

    Returns:
        list: Pixel offset defined in submask.
    """
    return params['d']

def __write_casa_region(region_object: 'astropy.region', coord:str, polygon_shape: str)->list:
    ''' Convert astropy region to casa region string. 

    Parameters
    ----------
    region_objects: astropy region
        Region definition in pixel or world coordinates
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel', 'world'.
    polygon_shape: str
        Region shape definition: rect: centerbox/rotbox, poly:poly 

    '''
    
    meta_value = []
    for key, value in region_object.meta.items():
        if key == 'corr':
            meta_value.append('{key}=[{pol}]'.format(key=key, pol=value))
        elif key == 'range':
            meta_value.append('{key}=[{lower}chan, {upper}chan]'.format(key=key, lower=str(value[0]), upper=str(value[1])))
        else:
            print('Unknown key: {} skipping'.format(key))
        
    meta = ', '.join(meta_value)

    if polygon_shape=='rect':
        if coord=='pixel':
            return 'centerbox[[{x} pix, {y} pix], [{width} pix, {height} pix]], {meta}'.format(
                x=region_object.center.x,
                y=region_object.center.y,
                width=region_object.width,
                height=region_object.height,
                meta=meta
            )
  
        if coord=='world':
            return 'rotbox[[{ra}, {dec}], [{width}, {height}], {angle}], {meta}'.format(
                ra=region_object.center.ra,
                dec=region_object.center.dec,
                width=region_object.width,
                height=region_object.height,
                angle=region_object.angle,
                meta=meta
            )
  
        else:
            raise RuntimeError('Unknown coordinate value: {}'.format(coord))
  
    if polygon_shape=='poly':    
        if coord=='pixel':
            coords_array = []
            for value in region_object.vertices:
                coords_array.append('[{x} pix, {y} pix]'.format(x=value.x, y=value.y))

            coords = ', '.join(coords_array)

            return 'poly[{coords}], {meta}'.format(coords=coords, meta=meta)

        if coord=='world':
            coords_array = []
            for value in region_object.vertices:
                coords_array.append('[{ra}, {dec}]'.format(ra=value.ra, dec=value.dec))

            coords = ', '.join(coords_array)

            return 'poly[{coords}], {meta}'.format(coords=coords, meta=meta)
        else:
            raise RuntimeError('Unknown coordinate value: {}'.format(coord))
    else:
        raise RuntimeError('Unknown polygon shape: {}'.format(polygon_shape))

def convert_masks(masks: dict, coord='pixel', cdesc=None)->list:
    '''Convert masks in standard format (as defined by ``CubeMask.jsmask_to_raw``) into
    other formats like list of CRTF, single region, etc.

    Parameters
    ----------
    masks: dict
        Dictionary containing ``masks`` and ``polys`` keys. The values for ``masks`` are the polygon
        references for each channel and the values for ``polys`` contain the points making up each
        polygon.
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel'.
    cdesc: dict
        Dictionary containing ``csys`` and ``shape`` which describes the coordinate system to be
        used for creating world coordinate coordinates. The ``shape`` is required along with the
        coordinate system for coordinate conversion.
    '''

    if cdesc is None or 'csys' not in cdesc or 'shape' not in cdesc:
        raise RuntimeError('region operations requires a coordinate description (cdesc parameter)')
    else:
        if not isinstance( cdesc['shape'], tuple ):
            raise RuntimeError("coordinate description must contain a 'shape' element of type 'tuple'")

        if isinstance( cdesc['csys'], dict ):
            csys = cdesc['csys']


    full_mask_params = []

    # Get the wcs object of the request format is world
    if coord == 'world':
        wcs = __get_wcs_object(cdesc['csys'])

    # Build list containing masks with channel and polaization information
    # in a more accesible way for the grouping funciton
    for index, mask in masks['masks'].items():
        for sub_mask in mask:
            sub_mask['s'], sub_mask['c'] = index
            full_mask_params.append(sub_mask)

    region_list = []

    # Group masks by unique center pixel and determine channel range. Extract 
    # mask properties and build astropy region for each center pixel and geometry
    # including region meta data.
    for key, props in itertools.groupby(full_mask_params, __get_center_pixels):
        channel_range = []
        for prop in props:
            channel_range.append(prop['c'])
            poly = prop['p']
            center_pixels = prop['d']
      
            mask_shape = masks['polys'][poly]['type']
      
            xs = masks['polys'][poly]['geometry']['xs']
            ys = masks['polys'][poly]['geometry']['ys']

            width = max(xs) - min(xs)
            height = max(ys) - min(ys)

            stokes_index = prop['s']
    
        if mask_shape=='rect':
            region = RectanglePixelRegion(
                PixCoord(x=center_pixels[0], y=center_pixels[1]), 
                width=width, height=height, angle=0*u.deg
        )
        elif mask_shape=='poly':
            region = PolygonPixelRegion(
                vertices=PixCoord(x=xs, y=ys)
        )

        else:
            raise RuntimeError('Invalid mask shape: {func}-->mask={mask}'.format(func=__convert_masks.__qualname__, mask=mask_shape))

        if coord=='world':
            region = region.to_sky(wcs)
            region.meta = {
                'corr': __index_to_stokes(stokes_index),
                'range': [min(channel_range), max(channel_range)]
            }

            region_list.append(__write_casa_region(region, coord=coord, polygon_shape=mask_shape))
        if coord=='pixel':

            region.meta = {
                'corr': __index_to_stokes(stokes_index),
                'range': [min(channel_range), max(channel_range)]
            }

            region_list.append(__write_casa_region(region, coord=coord, polygon_shape=mask_shape))

    return region_list

def __convert_masks(masks, format='crtf', coord='pixel', ret_type='str', cdesc=None):
    '''Convert masks in standard format (as defined by ``CubeMask.jsmask_to_raw``) into
    other formats like list of CRTF, single region, etc.

    Parameters
    ----------
    masks: dict
        Dictionary containing ``masks`` and ``polys`` keys. The values for ``masks`` are the polygon
        references for each channel and the values for ``polys`` contain the points making up each
        polygon.
    format: str
        Format string that indicates the format that should be returned. Allowed values are
        'region' or 'crtf'.
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel'.
    ret_type: str
        Data structure that should be returned the allowed values are 'str', 'list', 'singleton'
    cdesc: dict
        Dictionary containing ``csys`` and ``shape`` which describes the coordinate system to be
        used for creating world coordinate coordinates. The ``shape`` is required along with the
        coordinate system for coordinate conversion.
    '''
    if format == 'region':
        if __have_casatools == False:
            raise RuntimeError('casatools is not available, cannot create CASA regions')
        if cdesc is None or 'csys' not in cdesc or 'shape' not in cdesc:
            raise RuntimeError('region operations requires a coordinate description (cdesc parameter)')
        else:
            if not isinstance(cdesc['shape'], tuple):
                raise RuntimeError("coordinate description must contain a 'shape' element of type 'tuple'")
            if isinstance(cdesc['csys'], dict):
                csys_dict = cdesc['csys']
            else:
                csys_dict = cdesc['csys'].torecord()

    unique_polygons = {}
    ###
    ### collect masks ordered by (poly_index,dx,dy)
    ###
    for index, mask in masks['masks'].items():
        for poly in mask:
            unique_poly = tuple([poly['p']] + poly['d'])
            if unique_poly in unique_polygons:
                unique_polygons[unique_poly] += [index]
            else:
                unique_polygons[unique_poly] = [index]

    polygon_definitions = {index: poly for index, poly in masks['polys'].items()}

    if format == 'region' and coord == 'pixel':
        rg = regionmanager()
        rg.setcoordinates(csys_dict)

        ###
        ### create a region list given the index of the polygon used, the x/y translation, and the
        ### channels the region should be found on
        ###
        def create_regions_rgn(poly_index, xlate, channels):
            def create_result(shape, points):
                return [rg.fromtext(
                    f"{shape}[{points}],range=[{chan_range[0]}chan,{chan_range[1]}chan],corr=[{','.join(index_to_stokes(expand_range_incl(stokes_range)))}]" ,
                    shape=list(cdesc['shape']))
                        for chan_range in contiguous_ranges([c[1] for c in channels]) for stokes_range in
                        contiguous_ranges([c[0] for c in channels])]

            if polygon_definitions[poly_index]['type'] == 'poly':
                poly_pts = ','.join([f"[{x + xlate[0]}pix,{y + xlate[1]}pix]"
                                     for x in polygon_definitions[poly_index]['geometry']['xs']
                                     for y in polygon_definitions[poly_index]['geometry']['ys']])
                return create_result('poly', poly_pts)
            if polygon_definitions[poly_index]['type'] == 'rect':
                xs = polygon_definitions[poly_index]['geometry']['xs']
                ys = polygon_definitions[poly_index]['geometry']['ys']
                box_pts = f"[{min(xs) + xlate[0]}pix,{min(ys) + xlate[1]}pix],[{max(xs) + xlate[0]}pix,{max(ys) + xlate[1]}pix]"
                return create_result('box', box_pts)

        ### flatten list of unique polygon regions
        result = list(chain.from_iterable(
            [create_regions_rgn(index[0], index[1:], channels) for index, channels in unique_polygons.items()]))
        if ret_type == 'singleton':
            if len(result) == 0:
                raise RuntimeError("no regions created")
            if len(result) == 1:
                return result[0]
            ###return dict(enumerate(result)))          ## see: CAS-13764
            return rg.makeunion(dict(map(lambda t: (str(t[0]), t[1]), list(enumerate(result)))))
        if ret_type == 'list':
            return result
        raise RuntimeError(f"unknown ret_type for region format ({ret_type})")

    if format == 'crtf' and coord == 'pixel':
        ###
        ### create a region list given the index of the polygon used, the x/y translation, and the
        ### channels the region should be found on
        ###
        def create_regions_crt(poly_index, xlate, channels):
            def create_result(shape, points):
                return [
                    f"{shape}[{points}],range=[{chan_range[0]}chan,{chan_range[1]}chan],corr=[{','.join(index_to_stokes(expand_range_incl(stokes_range)))}]"
                    for chan_range in contiguous_ranges([c[1] for c in channels]) for stokes_range in
                    contiguous_ranges([c[0] for c in channels])]

            if polygon_definitions[poly_index]['type'] == 'poly':
                poly_pts = ','.join([f"[{x + xlate[0]}pix,{y + xlate[1]}pix]"
                                     for x in polygon_definitions[poly_index]['geometry']['xs']
                                     for y in polygon_definitions[poly_index]['geometry']['ys']])
                return create_result('poly', poly_pts)
            if polygon_definitions[poly_index]['type'] == 'rect':
                xs = polygon_definitions[poly_index]['geometry']['xs']
                ys = polygon_definitions[poly_index]['geometry']['ys']
                box_pts = f"[{min(xs) + xlate[0]}pix,{min(ys) + xlate[1]}pix],[{max(xs) + xlate[0]}pix,{max(ys) + xlate[1]}pix]"
                return create_result('box', box_pts)

        ###
        ### generate CFTF
        ###
        result = [create_regions_crt(index[0], index[1:], channels) for index, channels in unique_polygons.items()]
        if ret_type == 'str':
            return '\n'.join(chain.from_iterable(result))
        if ret_type == 'list':
            return list(chain.from_iterable(result))
        raise RuntimeError(f"unknown ret_type for crtf format ({ret_type})")
    raise RuntimeError(f"invalid format ({format}), or coord ({coord})")


def set_attributes(obj, **kw):
    '''Given an object and a set of keyword arguments, set the attributes
    in the object that correspond to the keywords to the specified values.

    Parameters
    ----------
    obj: object
        Object whose attributes should be set
    kw: keyword and object
        Attributes to be set

    Returns
    -------
    object
        ``obj`` parameter
    '''
    for k, v in kw.items():
        if hasattr(obj,k): setattr(obj, k, v)
    return obj


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            if get_ipython().__class__.__module__ == 'google.colab._shell':
                return True   # Google Colab
            else:
              return False  # Other type (?)
    except NameError:
        return False



########################################################################
#
# Copyright (C) 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''This contains functions which return the URLs to the ``casagui``
JavaScript libraries. The ``casalib`` library has Bokeh independent
functions while the `casaguijs` library has the Bokeh extensions'''
from ...utils import static_vars
from os import path

@static_vars( library_path=None )
def casalib_url( ):
    if casalib_url.library_path is None:
        casalib_url.library_path = path.join(path.dirname(path.dirname(path.dirname(__file__))), '__js__', 'casalib.min.js')
    if not path.isfile(casalib_url.library_path):
        raise RuntimeError( f''''casalib' JavaScript library not found at '{casalib_url.library_path}\'''' )
    return f'''file://{casalib_url.library_path}'''

@static_vars( library_path=None )
def casaguijs_url( ):
    if casaguijs_url.library_path is None:
        casaguijs_url.library_path = path.join(path.dirname(path.dirname(path.dirname(__file__))), '__js__', 'casaguijs.min.js')
    if not path.isfile(casaguijs_url.library_path):
        raise RuntimeError( f''''casaguijs' JavaScript library not found at '{casaguijs_url.library_path}\'''' )
    return f'''file://{casaguijs_url.library_path}'''



from enum import Enum

class DocEnum(Enum):
    '''Allows for documenting individual enum values::

           In [6]: class MaskMode(DocEnum):
           ...:     'Different masking modes available in addition to a user supplied mask'
           ...:     PB = 1, 'primary beam mask'
           ...:     AUTOMT = 2, 'multi-threshold auto masking'
           ...:

           In [7]: MaskMode.PB?
           Type:            MaskMode
           String form:     MaskMode.PB
           Docstring:       primary beam mask
           Class docstring: Different masking modes available in addition to a user supplied mask

           In [8]:
    '''
    def __new__(cls, value, doc=None):
        self = object.__new__(cls)  # calling super().__new__(value) here would fail
        self._value_ = value
        if doc is not None:
            self.__doc__ = doc
        return self



########################################################################
#
# Copyright (C) 2021,2022,2023,2024
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: aips2-request@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Implementation of ``ImagePipe`` class which provides a ``websockets``
implementation for CASA images which allows for interacitve display
of image cube channels in response to user input.'''

import os
import sys
import json
import asyncio
from uuid import uuid4

from . import DataPipe
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Tuple, String, Int, Instance, Nullable
from bokeh.models.callbacks import Callback
from bokeh.plotting import ColumnDataSource
from ..state import casalib_url, casaguijs_url

import numpy as np
try:
    import casatools as ct
    from casatools import regionmanager
    from casatools import image as imagetool
except:
    ct = None
    from casagui.utils import warn_import
    warn_import('casatools')

from ...utils import pack_arrays, partition, resource_manager, strip_arrays

class ImagePipe(DataPipe):
    """The `ImagePipe` allows for updates to Bokeh plots from a CASA or CNGI

    image. This is done using a `websocket`. A `ImagePipe` is created with
    the path to the image, and then it is used as the input to an
    `ImageDataSource` or a `SpectraDataSource`. This allows a single CASA
    or CNGI imge to be opened once and shared among multiple Bokeh plots,
    for example ploting an image channel and a plot of a spectra from the
    image cube.

    Attributes
    ----------
    address: tuple of string and int
        the string is the IP address for the network that should be used and the
        integer is the port number, see ``casagui.utils.find_ws_address``
    init_script: JavaScript
        this javascript is run when this DataPipe object is initialized. init_script
        is used to run caller JavaScript which needs to be run at initialization time.
        This is optional and does not need to be set.
    """
    __im_path = None
    __im = None
    __chan_shape = None

    shape = Tuple( Int, Int, Int, Int, help="shape: [ RA, DEC, Stokes, Spectral ]" )
    dataid = String( )
    fits_header_json = Nullable( String, help="""JSON representation of image FITS header for world coordinate labeling""" )
    _histogram_source = Nullable(Instance(ColumnDataSource), help='''
    data source for (raw) image channel histogram of intensities used with a "figure.quad(...)"
    ''')

    __javascript__ = [ casalib_url( ), casaguijs_url( ) ]

    def __open_image( self, image ):
        if self.__img is not None:
            self.__img.close( )
            self.__stokes_labels = None
        self.__img = imagetool( )
        self.__rgn = regionmanager( )
        try:
            self.__img.open(image)
            self.__image_path = image
        except Exception as ex:
            self.__img = None
            self.__image_path = None
            raise RuntimeError(f'could not open image: {image}') from ex
        imshape = self.__img.shape( )
        if self.__msk is not None and all(self.__msk.shape( ) != imshape):
            raise RuntimeError(f'mismatch between image shape ({imshape}) and mask shape ({self.__msk.shape( )})')
        if self.__chan_shape is None: self.__chan_shape = list(imshape[0:2])

    def __open_mask( self, mask ):
        if mask is None:
            self.__mask_path = None
            return
        if self.__msk is not None:
            self.__msk.close( )
        self.__msk = imagetool( )
        try:
            self.__msk.open(mask)
            self.__mask_path = mask
        except Exception as ex:
            self.__msk = None
            self.__mask_path = None
            raise RuntimeError(f'could not open mask: {mask}') from ex
        mskshape = self.__msk.shape( )
        if self.__img is not None and all(self.__img.shape( ) != mskshape):
            raise RuntimeError(f'mismatch between image shape ({self.__img.shape( )}) and mask shape ({mskshape})')
        if self.__chan_shape is None: self.__chan_shape = list(mskshape[0:2])

    def __close_mask( self ):
        if self.__msk is not None:
            self.__msk.close( )
            self.__msk = None

    def pixel_value( self, chan, index ):
        channel = self.__get_chan(chan)
        index[0] = min( index[0], channel.shape[0] - 1 )
        index[1] = min( index[1], channel.shape[1] - 1 )
        index[0] = max( index[0], 0 )
        index[1] = max( index[1], 0 )
        return np.squeeze(channel[index[0],index[1]])

    def stokes_labels( self ):
        """Returns stokes plane labels"""
        if self.__stokes_labels is None:
            self.__stokes_labels = self.__img.coordsys( ).stokes( )
        return self.__stokes_labels

    def __get_chan( self, index ):
        def newest_ctime( path ):
            files = os.listdir(path)
            paths = [os.path.join(path, basename) for basename in files]
            return max( map( os.path.getctime, paths ) )

        image_ctime = newest_ctime( self.__image_path )
        if image_ctime > self.__cached_chan_ctime or \
           self.__cached_chan_index[0] != index[0] or \
           self.__cached_chan_index[1] != index[1] or \
           self.__cached_chan is None :
            if self.__img is None:
                raise RuntimeError('no image is available')
            ###
            ### ensure that the channel index is within cube shape
            ###
            index = list(index)     # index is potentially a python tuple
            index[0] = min( index[0], self.__chan_shape[0] - 1 )
            index[1] = min( index[1], self.__chan_shape[1] - 1 )
            index[0] = max( index[0], 0 )
            index[1] = max( index[1], 0 )
            self.__cached_chan_index = index
            self.__cached_chan_ctime = image_ctime
            self.__cached_chan = self.__img.getchunk( blc=[0,0] + index,
                                                      trc=self.__chan_shape + index )
        return self.__cached_chan

    ### ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----
    ### the element type of image pixels retrieved from the CASA image are float64, but it
    ### seems like 256 is the greatest number of colors in the colormaps currrently used
    ### for pseudo color within interactive clean...
    ### ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----
    def channel( self, index, pixel_type ):
        """Retrieve one channel from the image cube. The `index` should be a
        two element list of integers. The first integer is the ''stokes'' axis
        in the image cube. The second integer is the ''channel'' axis in the
        image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        pixel_type: numpy type
            the numpy type for the pixel elements of the returned channel
        """
        def quantize( nptype, image_plane ):
            lower_bin = np.array([],bool)
            upper_bin = np.array([],bool)
            mask = np.array([],bool)
            bits = nptype(0).nbytes * 8

            if self.__quant_adjustments['transfer']['scaling'] != 'linear':
                if self.__quant_adjustments['transfer']['scaling'] not in self.__quant_scaling:
                    print( f'''error: ${self.__quant_adjustments['transfer']['scaling']} is not a known scaling...''', file=sys.stderr )
                    scaled_plane = image_plane
                else:
                    scaled_plane = self.__quant_scaling[self.__quant_adjustments['transfer']['scaling']](image_plane,**self.__quant_adjustments['transfer']['args'])
            else:
                scaled_plane = image_plane

            #####################################################################################
            ### The assumption is that np.uint8 implies a pseudo color mapping. When this     ###
            ### type is used the quantization tweaks provided via 'adjust-colormap' are used. ###
            #####################################################################################
            if nptype == np.uint8 and \
               ( len(self.__quant_adjustments['bounds'][0]) > 0 or \
                 len(self.__quant_adjustments['bounds'][1]) > 0 ):
                ################################################################################
                ### [ ] means that the user has not modified that boundary                   ###
                ################################################################################
                lower_bin = np.array([],bool) if len(self.__quant_adjustments['bounds'][0]) == 0 else image_plane <= self.__quant_adjustments['bounds'][0][0]
                upper_bin = np.array([],bool) if len(self.__quant_adjustments['bounds'][1]) == 0 else image_plane >= self.__quant_adjustments['bounds'][1][0]

                if lower_bin.any( ) and upper_bin.any( ):
                    mask = ~( lower_bin | upper_bin )
                elif lower_bin.any( ):
                    mask = ~lower_bin
                elif upper_bin.any( ):
                    mask = ~upper_bin

                if mask.any( ):
                    min = scaled_plane[mask].min( )
                    max = scaled_plane[mask].max( )
                    max -= min
                    result = np.ma.zeros(scaled_plane.shape,scaled_plane.dtype)
                    if upper_bin.any( ):
                        result[mask] = ((scaled_plane[mask] - min)/max) * (2**bits-3) + 1
                        result[upper_bin] = 255
                    else:
                        result[mask] = ((scaled_plane[mask] - min)/max) * (2**bits-1)
                    result = result.filled(0).astype(np.uint8)
                    return result

            min = scaled_plane.min( )
            max = scaled_plane.max( )
            max -= min
            img = ((scaled_plane - min)/max) * (2**bits-1)
            result = img.astype(nptype)
            return result

        if self.__img is None:
            raise RuntimeError('no image is available')
        if np.issubdtype( pixel_type, np.integer ):
            return quantize( pixel_type,
                             np.squeeze( self.__get_chan(index) ) ).transpose( )
        else:
            return np.squeeze( self.__get_chan(index) ).astype(pixel_type).transpose( )

    def have_mask0( self ):
        """Check to see if the synthesis imaging 'mask0' mask exists

        Returns
        -------
        bool:
            ''True'' if the cube contains an internal ''mask0'' mask otherwise ''False''
        """
        if self.__img is None:
            raise RuntimeError('no image is available')
        return 'mask0' in self.__img.maskhandler('get')


    def mask0( self, index ):
        """Within the image, there can be an arbitrary number of INTERNAL masks. They can
        have arbitrary names. The synthesis imaging module uses a mask named 'mask0'. This
        mask is used in processing (it MAY represent the beam).

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        """
        ### ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----
        ### tclean does not maintain mask0. Instead, calls to tclean can result in the
        ### internal, mask0 being lost. Because of this, once a good copy of this internal
        ### mask is retrieved it is reused. Urvashi says that reusing one copy throughout
        ### should be fine (Fri Mar 31 13:54:17 EDT 2023)
        ### ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----
        if self.__img is None:
            raise RuntimeError('no image is available')
        if self.__mask0_cache is None and self.have_mask0( ):
            self.__mask0_cache = self.__img.getregion(getmask=True)
        if self.__mask0_cache is not None:
            return self.__mask0_cache[:,:,index[0],index[1]]
        return None

    def have_mask( self ):
        """Check to see if a mask exists.

        Returns
        -------
        bool:
            ''True'' if a mask cube is available otherwise ''False''
        """
        return self.__msk is not None

    def mask( self, index, modify=False ):
        """Retrieve one channel mask from the mask cube. The `index` should be a
        two element list of integers. The first integer is the ''stokes'' axis
        in the image cube. The second integer is the ''channel'' axis in the
        image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        modify: boolean
            If true, it implies that the channel mask is being retrieved for modification
            and updating the channel on disk. If false, it implies that the channel mask
            is being retrieved for display.
        """
        if self.__msk is None:
            raise RuntimeError(f'cannot retrieve mask at {repr(index)} because no mask cube exists')
        return ( np.squeeze( self.__msk.getchunk( blc=[0,0] + index,
                                                trc=self.__chan_shape + index) ).astype(np.bool_).transpose( )
                 if modify == False else
                 np.squeeze( self.__msk.getchunk( blc=[0,0] + index,
                                                trc=self.__chan_shape + index) ) )

    def mask_value( self, chan, index ):
        if self.__msk is None:
            raise RuntimeError(f'cannot retrieve mask at {repr(index)} because no mask cube exists')
        try:
            pv = self.__msk.pixelvalue( index + chan )
            return int(pv['value']['value'])
        except:
            pass
        return -1

    def set_mask_name( self, new_mask_path ):
        self.__close_mask( )
        self.__open_mask( new_mask_path )

    def put_mask( self, index, mask ):
        """Replace one channel mask with the mask specified as the second parameter.
        The `index` should be a two element list of integers. The first integer is the
        ''stokes'' axis in the image cube. The second integer is the ''channel'' axis
        in the image cube. The assumption is that the :code:`mask` parameter was retrieved
        from the mask cube using the :code:`mask(...)` function with the :code:`modify`
        parameter set to :code:`True`.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        mask: numpy.ndarray
            two dimensional array to replace the existing mask for the channel specified
            by :code:`index`
        """
        if self.__msk is None:
            raise RuntimeError(f'cannot replace mask at {repr(index)} because no mask cube exists')
        if mask.dtype == bool:
            ### cannot put bools with putchunk
            self.__msk.putchunk( blc=[0,0] + index, pixels=mask.astype(np.uint8) )
        else:
            self.__msk.putchunk( blc=[0,0] + index, pixels=mask )

    def spectra( self, index ):
        """Retrieve one spectra from the image cube. The `index` should be a
        three element list of integers. The first integer is the ''right
        ascension'' axis, the second integer is the ''declination'' axis,
        and the third integer is the ''stokes'' axis.

        Parameters
        ----------
        index: [ int, int, int ]
            list containing first the ''right ascension'', the ''declination'' and
            the ''stokes'' axis
        """
        index = list(map( lambda i: 0 if i is None else i, index ))
        if index[0] >= self.shape[0]:
            index[0] = self.shape[0] - 1
        if index[1] >= self.shape[1]:
            index[1] = self.shape[1] - 1
        if self.__img is None:
            raise RuntimeError('no image is available')
        result = np.squeeze( self.__img.getchunk( blc=index + [0],
                                                 trc=index + [self.shape[-1]] ) )
        ### should return spectral freq etc.
        ### here for X rather than just the index
        try:
            return { 'x': list(range(len(result))), 'y': list(result) }
        except Exception:
            ## In this case, result is not iterable (e.g.) only one channel in the cube.
            ## A zero length numpy ndarray has no shape and looks like a float but it is
            ## an ndarray.
            return { 'x': [0], 'y': [float(result)] }

    def histogram_source( self, data ):
        if not self._histogram_source:
            self._histogram_source = ColumnDataSource( data=data )
        return self._histogram_source

    async def _image_message_handler( self, cmd ):
        if cmd['action'] == 'channel':
            chan = self.channel(cmd['index'],np.uint8)
            mask = { } if self.__msk is None else { 'msk': [ pack_arrays( self.mask(cmd['index']) ) ] }
            _mask0 = self.mask0(cmd['index'])
            mask0 = { } if _mask0 is None else { 'msk0': [ pack_arrays(_mask0) ] }
            histogram = self.histogram( cmd['index'] ) if self._histogram_source else { }
            if self._stats:
                #statistics for the displayed plane of the image cubea
                statistics = self.statistics( cmd['index'] )
                return { 'chan': { 'img': [ pack_arrays(chan) ],
                                   **mask0,
                                   **mask },
                         'stats': { 'labels': list(statistics.keys( )), 'values': pack_arrays(list(statistics.values( ))) },
                         'hist': histogram,
                         'id': cmd['id'] }
            else:
                return { 'chan': { 'img': [ pack_arrays(chan) ],
                                   **mask0,
                                   **mask },
                         'hist': histogram,
                         'id': cmd['id'] }

        elif cmd['action'] == 'spectra':
            return { 'spectrum': pack_arrays( self.spectra(cmd['index']) ), 'id': cmd['id'] }
        elif cmd['action'] == 'adjust-colormap':
            if cmd['bounds'] == "reset":
                self.__quant_adjustments = { 'bounds': [ [ ], [ ] ],
                                             'transfer': {'scaling': 'linear'} }
            else:
                ### later a function should be provided for setting the quantization transfer function
                self.__quant_adjustments = { 'bounds': cmd['bounds'], 'transfer': cmd['transfer'] }
                ### ensure that the cached channel is not used...
                self.__cached_chan = None
            return { 'result': 'OK', 'id': cmd['id'] }

    def __init__( self, image, *args, mask=None, stats=False, **kwargs ):
        super( ).__init__( *args, **kwargs, )

        self.dataid = str(uuid4( ))

        if ct is None:
            raise RuntimeError('cannot open an image because casatools is not available')

        self.__img = None
        self.__msk = None
        resource_manager( ).reg_at_exit( self, '__del__' )
        self._stats = stats
        self.__open_image( image )
        self.__open_mask( mask )
        self.__mask0_cache = None
        self.shape = list(self.__img.shape( ))
        if not self.fits_header_json:
            __hdr_dict = self.__img.fitsheader(exclude="HISTORY")
            if __hdr_dict:
                self.fits_header_json = json.dumps(strip_arrays(__hdr_dict))
        self.__session = None
        self.__stokes_labels = None
        self.__mask_statistics = False

        ###
        ### the last channel retrieved is kept around for pixel retrieval
        ###
        self.__cached_chan = None
        self.__cached_chan_index = None
        self.__cached_chan_ctime = 0

        ###
        ### quantization controls to affect how pseudo colors are displayed
        ###
        self.__quant_adjustments = { 'bounds': [ [ ], [ ] ],
                                     'transfer': {'scaling': 'linear'} }
        self.__quant_scaling = { 'log':    lambda chan,alpha: np.ma.log(alpha * chan + 1.0) / np.ma.log(alpha + 1.0),
                                 'sqrt':   lambda chan:       np.ma.sqrt(chan),
                                 'square': lambda chan:       np.square(chan),
                                 'gamma':  lambda chan,gamma: np.ma.power(chan,gamma),
                                 'power':  lambda chan,alpha: (np.ma.power(alpha,chan) - 1.0) / alpha }

        super( ).register( self.dataid, self._image_message_handler )

    def __del__(self):
        if self.__rgn:
            self.__rgn.done( )
        if self.__img != None:
            self.__img.close()
            self.__img.done()
            self.__img = None
            self.__stokes_labels = None

    def coorddesc( self ):
        ia = imagetool( )
        ia.open(self.__image_path)
        csys = ia.coordsys( )
        ia.close( )
        return { 'csys': csys, 'shape': tuple(self.shape) }

    def statistics_config( self, use_mask=None ):
        '''Configure the behavior of the statistics function.
        use_mask indicates that if a mask is available, the statistics should
        be based upon the portion of the image included in the mask instead of
        the whole channel.'''
        if self.__mask_path and use_mask is not None:
            self.__mask_statistics = bool(use_mask)

    def statistics( self, index ):
        """Retrieve statistics for one channel from the image cube. The `index`
        should be a two element list of integers. The first integer is the
        ''stokes'' axis in the image cube. The second integer is the ''channel''
        axis in the image cube.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        """
        def singleton( potential_nonlist ):
            # convert a list of a single element to the element
            return potential_nonlist if len(potential_nonlist) != 1 else potential_nonlist[0]
        def sort_result( unsorted_dictionary ):
            part = partition( lambda s: (s.startswith('trc') or s.startswith('blc')), sorted(unsorted_dictionary.keys( )) )
            return { k: unsorted_dictionary[k] for k in part[1] + part[0] }

        reg = self.__rgn.box( [0,0] + index, self.__chan_shape + index )
        ###
        ### This seems like it should work:
        ###
        #      rawstats = self.__img.statistics( region=reg )
        ###
        ### but it does not so we have to create a one-use image tool (see CAS-13625)
        ###
        ia = imagetool( )
        ia.open(self.__image_path)
        if self.__mask_statistics:
            rawstats = ia.statistics( region=reg, mask=self.__mask_path )
        else:
            rawstats = ia.statistics( region=reg )
        ia.close( )
        return sort_result( { k: singleton([ x.item( ) for x in v ]) if isinstance(v,np.ndarray) else v for k,v in rawstats.items( ) } )

    def histogram( self, index ):
        """Calculate histogram (Bokeh Quad) extents for update of colormap adjuster (or anything
        else that wants a histogram of image intensities.

        Parameters
        ----------
        index: [ int, int ]
            list containing first the ''stokes'' index and second the ''channel'' index
        """
        if not self._histogram_source:
            return { }

        chan = self.__get_chan(index)
        bins = np.linspace( chan.min( ), chan.max( ), len(self._histogram_source.data['top'])+1 )
        hist, edges = np.histogram( chan, density=False, bins=bins )
        return dict( left=list(edges[:-1]), right=list(edges[1:]), top=list(hist), bottom=[0]*len(hist) )



#from casagui.bokeh.state import initialize_bokeh
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.1.min.js",
#                  "../../../casaguijs/dist/casaguijs.js" )                        ### local build
#initialize_bokeh( "../../../casalib/dist/esbuild/casalib-v0.0.2.min.js" )
#
#initialize_bokeh( bokeh='../../../casaguijs/dist/casaguijs.js' )                  ### local gui/js build + standard bokeh independent library

import os
import ssl
import certifi
import asyncio
import urllib
import tarfile

from casagui.apps import InteractiveClean

##
## demo measurement set to use
##
ms_path = 'uranus_averaged_380MHzchan_6stime.ms'
##
## where to fetch the demo measurement set
##
ms_url = "https://casa.nrao.edu/download/devel/casavis/data/uranus_averaged_380MHzchan_6stime-ms.tar.gz"
##
## output image file name
##
img = 'uranus'

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall( )
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

if not os.path.isdir(ms_path):
    raise  RuntimeError("Failed to fetch measurement set")

ic = InteractiveClean( vis=ms_path, imagename=img,
                       field='0', spw='',
                       gridder='standard',
                       imsize=2048,
                       cell='0.08arcsec',
                       specmode='mfs',
                       interpolation='nearest',
                       weighting='briggs',
                       pblimit=-1e-05,
                       deconvolver='hogbom',
                       threshold='0.001Jy',
                       niter=500,
                       cycleniter=10,
                       cyclefactor=3 )

if True:
    print( "Result: %s" % ic( ) )
    print( " Masks: %s" % repr(ic.masks( )) )
else:
    try:
        asyncio.get_event_loop().run_until_complete(ic.show( ))
        asyncio.get_event_loop().run_forever()

    except KeyboardInterrupt:
        print('\nInterrupt received, shutting down ...')
        #os.system('rm -rf {output_image}.* *.html *.log'.format(output_image=output_image))

    print( "Result: %s" % repr(ic.result( )) )
    print( " Masks: %s" % repr(ic.masks( )) )
